//block_array.cc: Implementation file for block_array.h

#include "signal.h"
#include "cell.h"
#include "blocks.h"
#include "block_array.h"
#include <iostream.h>
#include <iomanip.h>
#include <string.h>

///////////////////////////////////////////////////////////////////////////////

//FIXME: Use structs of names and numbers
//Routines for going both ways using structs (hash function?)
void Pin_Name_To_String(pin_name_type pin_name, char* string)
{
  switch(pin_name)
    {
    case L5NIN_H:   strcpy(string, "L5NIN_H");      break;
    case S1OUT_H:   strcpy(string, "S1OUT_H");      break;
    case S1NIN_H:   strcpy(string, "S1NIN_H");      break;
    case S1WIN_H:   strcpy(string, "S1WIN_H");      break;
    case PAD_VDD1:  strcpy(string, "PAD_VDD1");     break;
    case SCANIN_H:  strcpy(string, "SCANIN_H");     break;
    case RST_L:     strcpy(string, "RST_L");        break;
    case L1WIN_H:   strcpy(string, "L1WIN_H");      break;
    case L1OUT_H:   strcpy(string, "L1OUT_H");      break;
    case GND:       strcpy(string, "GND");          break;
    case CLK_A:     strcpy(string, "CLK_A");        break;
    case L0WIN_H:   strcpy(string, "L0WIN_H");      break;
    case L0OUT_H:   strcpy(string, "L0OUT_H");      break;
    case CONFOUT_H: strcpy(string, "CONFOUT_H");    break;
    case PAD_VDD2:  strcpy(string, "PAD_VDD2");     break;
    case S0WIN_H:   strcpy(string, "S0WIN_H");      break;
    case S0OUT_H:   strcpy(string, "S0OUT_H");      break;
    case S0SIN_H:   strcpy(string, "S0SIN_H");      break;
    case SCANOUT_H: strcpy(string, "SCANOUT_H");    break;
    case L2SIN_H:   strcpy(string, "L2SIN_H");      break;
    case L2OUT_H:   strcpy(string, "L2OUT_H");      break;
    case S2OUT_H:   strcpy(string, "S2OUT_H");      break;
    case S2SIN_H:   strcpy(string, "S2SIN_H");      break;
    case CONFIN_H:  strcpy(string, "CONFIN_H");     break;
    case PAD_GND1:  strcpy(string, "PAD_GND1");     break;
    case TEST_H:    strcpy(string, "TEST_H");       break;
    case S2EIN_H:   strcpy(string, "S2EIN_H");      break;
    case L6EIN_H:   strcpy(string, "L6EIN_H");      break;
    case L6OUT_H:   strcpy(string, "L6OUT_H");      break;
    case VDD:       strcpy(string, "VDD");          break;
    case CLK_B:     strcpy(string, "CLK_B");        break;
    case L7EIN_H:   strcpy(string, "L7EIN_H");      break;
    case L7OUT_H:   strcpy(string, "L7OUT_H");      break;
    case CONFOPT_H: strcpy(string, "CONFOPT_H");    break;
    case PAD_GND2:  strcpy(string, "PAD_GND2");     break;
    case ENCONF_H:  strcpy(string, "ENCONF_H");     break;
    case S3EIN_H:   strcpy(string, "S3EIN_H");      break;
    case S3OUT_H:   strcpy(string, "S3OUT_H");      break;
    case S3NIN_H:   strcpy(string, "S3NIN_H");      break;
    case L5OUT_H:   strcpy(string, "L5OUT_H");      break;
    default:        strcpy(string, "UNDEFINED");
    } //switch
} //Pin_Name_To_String


pin_name_type String_To_Pin_Name(char* string)
{
  unsigned int index;
  char current_name[PIN_NAME_SIZE];
  for(index = 0; index<PINS; index++)
    {
      Pin_Name_To_String((pin_name_type)index, current_name);
      if(strcasecmp(current_name, string)==0) 
	return index;
    }
  return UNDEFINED_PIN;
} //String_To_Pin_Name

///////////////////////////////////////////////////////////////////////////////


pin_name_type String_To_CSB_Name(char* string)
{
  if(strcasecmp("CSB0", string)==0) return CSB0;
  if(strcasecmp("CSB1", string)==0) return CSB1;
  if(strcasecmp("CSB2", string)==0) return CSB2;
  if(strcasecmp("CSB3", string)==0) return CSB3;
  return UNDEFINED_CSB;
} //String_To_CSB_Name


///////////////////////////////////////////////////////////////////////////////

CROWBAR_class::CROWBAR_class()
{
  Clear(); //Clear data structures
  //Next configure pins
  Pin_list[L5NIN_H].Configure(dedicated_input, asserted_high);
  Pin_list[S1OUT_H].Configure(dedicated_output, asserted_high);
  Pin_list[S1NIN_H].Configure(dedicated_input, asserted_high); 
  Pin_list[S1WIN_H].Configure(dedicated_input, asserted_high);
  Pin_list[PAD_VDD1].Configure(vdd, asserted_high);

  Pin_list[SCANIN_H].Configure(dedicated_input, asserted_high);
  Pin_list[RST_L].Configure(dedicated_input, asserted_low);
  Pin_list[L1WIN_H].Configure(dedicated_input, asserted_high);
  Pin_list[L1OUT_H].Configure(dedicated_output, asserted_high);
  Pin_list[GND].Configure(gnd, asserted_high);

  Pin_list[CLK_A].Configure(dedicated_input, asserted_high);
  Pin_list[L0WIN_H].Configure(dedicated_input, asserted_high);
  Pin_list[L0OUT_H].Configure(dedicated_output, asserted_high);
  Pin_list[CONFOUT_H].Configure(dedicated_output, asserted_high);
  Pin_list[PAD_VDD2].Configure(vdd, asserted_high);

  Pin_list[S0WIN_H].Configure(dedicated_input, asserted_high);
  Pin_list[S0OUT_H].Configure(dedicated_output, asserted_high);
  Pin_list[S0SIN_H].Configure(dedicated_input, asserted_high);
  Pin_list[SCANOUT_H].Configure(dedicated_output, asserted_high);
  Pin_list[L2SIN_H].Configure(dedicated_input, asserted_high);

  Pin_list[L2OUT_H].Configure(dedicated_output, asserted_high);
  Pin_list[S2OUT_H].Configure(dedicated_output, asserted_high);
  Pin_list[S2SIN_H].Configure(dedicated_input, asserted_high);
  Pin_list[CONFIN_H].Configure(dedicated_input, asserted_high);
  Pin_list[PAD_GND1].Configure(gnd, asserted_high);

  Pin_list[TEST_H].Configure(dedicated_input, asserted_high);
  Pin_list[S2EIN_H].Configure(dedicated_input, asserted_high);
  Pin_list[L6EIN_H].Configure(dedicated_input, asserted_high);
  Pin_list[L6OUT_H].Configure(dedicated_output, asserted_high);
  Pin_list[VDD].Configure(vdd, asserted_high);

  Pin_list[CLK_B].Configure(dedicated_input, asserted_high);
  Pin_list[L7EIN_H].Configure(dedicated_input, asserted_high);
  Pin_list[L7OUT_H].Configure(dedicated_output, asserted_high);
  Pin_list[CONFOPT_H].Configure(dedicated_input, asserted_high);
  Pin_list[PAD_GND2].Configure(gnd, asserted_high);

  Pin_list[ENCONF_H].Configure(dedicated_input, asserted_high);
  Pin_list[S3EIN_H].Configure(dedicated_input, asserted_high);
  Pin_list[S3OUT_H].Configure(dedicated_output, asserted_high);
  Pin_list[S3NIN_H].Configure(dedicated_input, asserted_high);
  Pin_list[L5OUT_H].Configure(dedicated_output, asserted_high);
} //constructor


CROWBAR_class::~CROWBAR_class()
{
} //destructor


void CROWBAR_class::Clear()
{
  unsigned int index;
  for(index = 0; index<PINS; index++)
    Pin_list[index].Clear();
  for(index = 0; index<LOGIC_BLOCKS; index++)
    Sensor_block_list[index].Clear();
  for(index = 0; index<SENSOR_BLOCKS; index++)
    {
      Logic_block_list[index].Clear();
      image[index] = Undefined;
    } //for
} //clear array data structures


void CROWBAR_class::Update()
{
  unsigned int index;
  
  //Drive outputs onto pins (block outputs, scan-out, config-out)
  Apply_Signal_To_Pin(S0OUT_H, Sensor_block_list[0].Get_Output());
  Apply_Signal_To_Pin(S1OUT_H, Sensor_block_list[1].Get_Output());
  Apply_Signal_To_Pin(S2OUT_H, Sensor_block_list[2].Get_Output());
  Apply_Signal_To_Pin(S3OUT_H, Sensor_block_list[3].Get_Output());
  Apply_Signal_To_Pin(L0OUT_H, Logic_block_list[0].Get_Output());
  Apply_Signal_To_Pin(L1OUT_H, Logic_block_list[1].Get_Output());
  Apply_Signal_To_Pin(L2OUT_H, Logic_block_list[2].Get_Output());
  Apply_Signal_To_Pin(L5OUT_H, Logic_block_list[5].Get_Output());
  Apply_Signal_To_Pin(L6OUT_H, Logic_block_list[6].Get_Output());
  Apply_Signal_To_Pin(L7OUT_H, Logic_block_list[7].Get_Output());
  Apply_Signal_To_Pin(SCANOUT_H, Sensor_block_list[0].Get_Test_Output());
  Apply_Signal_To_Pin(CONFOUT_H, Logic_block_list[0].Get_Mux_Config_Out());
  
  //updates pads to get new inputs/outputs
  for(index = 0; index<PINS; index++)
    {
      Pin_list[index].Update();
    } //for

  //Get inputs to each block (this gets nasty)
  //I should define a more simple way to do this
  CLB_input CLB_inputs[LOGIC_BLOCKS];
  CSB_input CSB_inputs[SENSOR_BLOCKS];

  //some signals are global (easy)
  for(index = 0; index<LOGIC_BLOCKS; index++)
    {
      CLB_inputs[index].clk_a = Get_Pin_Signal(CLK_A);
      CLB_inputs[index].clk_b = Get_Pin_Signal(CLK_B);
      CLB_inputs[index].reset_flop = Get_Pin_Signal(RST_L);
      CLB_inputs[index].test_mode = Get_Pin_Signal(TEST_H);
      CLB_inputs[index].config_enable = Get_Pin_Signal(ENCONF_H);
    } //for

  for(index = 0; index<SENSOR_BLOCKS; index++)
    {
      CSB_inputs[index].clk_a = Get_Pin_Signal(CLK_A);
      CSB_inputs[index].clk_b = Get_Pin_Signal(CLK_B);
      CSB_inputs[index].reset_flop = Get_Pin_Signal(RST_L);
      CSB_inputs[index].test_mode = Get_Pin_Signal(TEST_H);
      CSB_inputs[index].config_enable = Get_Pin_Signal(ENCONF_H);
      CSB_inputs[index].config_opt = Get_Pin_Signal(CONFOPT_H);
    } //for

  //some signals form chains (scan-chain, config-chain)
  //scan
  CSB_inputs[1].scan_in = Get_Pin_Signal(SCANIN_H);
  CLB_inputs[5].scan_in = Sensor_block_list[1].Get_Test_Output();
  CSB_inputs[3].scan_in = Logic_block_list[5].Get_Test_Output();
  CLB_inputs[7].scan_in = Sensor_block_list[3].Get_Test_Output();
  CLB_inputs[4].scan_in = Logic_block_list[7].Get_Test_Output();
  CLB_inputs[1].scan_in = Logic_block_list[4].Get_Test_Output();
  CLB_inputs[0].scan_in = Logic_block_list[1].Get_Test_Output();
  CLB_inputs[3].scan_in = Logic_block_list[0].Get_Test_Output();
  CLB_inputs[6].scan_in = Logic_block_list[3].Get_Test_Output();
  CSB_inputs[2].scan_in = Logic_block_list[6].Get_Test_Output();
  CLB_inputs[2].scan_in = Sensor_block_list[2].Get_Test_Output();
  CSB_inputs[0].scan_in = Logic_block_list[2].Get_Test_Output();

  //config

  CSB_inputs[2].config_in = Get_Pin_Signal(CONFIN_H);
  CLB_inputs[6].lut_config_in = Sensor_block_list[2].Get_Config_Out();
  CLB_inputs[7].lut_config_in = Logic_block_list[6].Get_Lut_Config_Out();
  CSB_inputs[3].config_in = Logic_block_list[7].Get_Lut_Config_Out();
  CLB_inputs[7].mux_config_in = Sensor_block_list[3].Get_Config_Out();
  CLB_inputs[6].mux_config_in = Logic_block_list[7].Get_Mux_Config_Out();
  CLB_inputs[2].lut_config_in = Logic_block_list[6].Get_Mux_Config_Out();
  CLB_inputs[3].lut_config_in = Logic_block_list[2].Get_Lut_Config_Out();
  CLB_inputs[4].lut_config_in = Logic_block_list[3].Get_Lut_Config_Out();
  CLB_inputs[5].lut_config_in = Logic_block_list[4].Get_Lut_Config_Out();
  CLB_inputs[5].mux_config_in = Logic_block_list[5].Get_Lut_Config_Out();
  CLB_inputs[4].mux_config_in = Logic_block_list[5].Get_Mux_Config_Out();
  CLB_inputs[3].mux_config_in = Logic_block_list[4].Get_Mux_Config_Out();
  CLB_inputs[2].mux_config_in = Logic_block_list[3].Get_Mux_Config_Out();
  CSB_inputs[0].config_in = Logic_block_list[2].Get_Mux_Config_Out();
  CLB_inputs[0].lut_config_in = Sensor_block_list[0].Get_Config_Out();
  CLB_inputs[1].lut_config_in = Logic_block_list[0].Get_Lut_Config_Out();
  CSB_inputs[1].config_in = Logic_block_list[1].Get_Lut_Config_Out();
  CLB_inputs[1].mux_config_in = Sensor_block_list[1].Get_Config_Out();
  CLB_inputs[0].mux_config_in = Logic_block_list[1].Get_Mux_Config_Out();

  //others connect blocks to their neighbors (possibly pins)
  CLB_inputs[0].north = Logic_block_list[1].Get_Output(); 
  CLB_inputs[0].south = Sensor_block_list[0].Get_Output();
  CLB_inputs[0].east = Logic_block_list[3].Get_Output();
  CLB_inputs[0].west = Get_Pin_Signal(L0WIN_H);

  CLB_inputs[1].north = Sensor_block_list[1].Get_Output(); 
  CLB_inputs[1].south = Logic_block_list[0].Get_Output();
  CLB_inputs[1].east = Logic_block_list[4].Get_Output();
  CLB_inputs[1].west = Get_Pin_Signal(L1WIN_H);

  CLB_inputs[2].north = Logic_block_list[3].Get_Output();
  CLB_inputs[2].south = Get_Pin_Signal(L2SIN_H);
  CLB_inputs[2].east = Sensor_block_list[2].Get_Output();
  CLB_inputs[2].west = Sensor_block_list[0].Get_Output();

  CLB_inputs[3].north = Logic_block_list[4].Get_Output();
  CLB_inputs[3].south = Logic_block_list[2].Get_Output();
  CLB_inputs[3].east = Logic_block_list[6].Get_Output();
  CLB_inputs[3].west = Logic_block_list[0].Get_Output();

  CLB_inputs[4].north = Logic_block_list[5].Get_Output();
  CLB_inputs[4].south = Logic_block_list[3].Get_Output();
  CLB_inputs[4].east = Logic_block_list[7].Get_Output();
  CLB_inputs[4].west = Logic_block_list[1].Get_Output();
  
  CLB_inputs[5].north = Get_Pin_Signal(L5NIN_H);
  CLB_inputs[5].south = Logic_block_list[4].Get_Output();
  CLB_inputs[5].east = Sensor_block_list[3].Get_Output();
  CLB_inputs[5].west = Sensor_block_list[1].Get_Output();

  CLB_inputs[6].north = Logic_block_list[7].Get_Output();
  CLB_inputs[6].south = Sensor_block_list[2].Get_Output();
  CLB_inputs[6].east = Get_Pin_Signal(L6EIN_H);
  CLB_inputs[6].west = Logic_block_list[3].Get_Output();

  CLB_inputs[7].north = Sensor_block_list[3].Get_Output();
  CLB_inputs[7].south = Logic_block_list[6].Get_Output();
  CLB_inputs[7].east = Get_Pin_Signal(L7EIN_H);
  CLB_inputs[7].west = Logic_block_list[4].Get_Output();

  CSB_inputs[0].north = Logic_block_list[0].Get_Output();
  CSB_inputs[0].south = Get_Pin_Signal(S0SIN_H);
  CSB_inputs[0].east = Logic_block_list[2].Get_Output();
  CSB_inputs[0].west = Get_Pin_Signal(S0WIN_H);
  CSB_inputs[0].sensor = image[0];

  CSB_inputs[1].north = Get_Pin_Signal(S1NIN_H);
  CSB_inputs[1].south = Logic_block_list[1].Get_Output();
  CSB_inputs[1].east = Logic_block_list[5].Get_Output();
  CSB_inputs[1].west = Get_Pin_Signal(S1WIN_H);
  CSB_inputs[1].sensor = image[1];

  CSB_inputs[2].north = Logic_block_list[6].Get_Output();
  CSB_inputs[2].south = Get_Pin_Signal(S2SIN_H);
  CSB_inputs[2].east = Get_Pin_Signal(S2EIN_H);
  CSB_inputs[2].west = Logic_block_list[2].Get_Output();
  CSB_inputs[2].sensor = image[2];

  CSB_inputs[3].north = Get_Pin_Signal(S3NIN_H);
  CSB_inputs[3].south = Logic_block_list[7].Get_Output();
  CSB_inputs[3].east = Get_Pin_Signal(S3EIN_H);
  CSB_inputs[3].west = Logic_block_list[5].Get_Output();
  CSB_inputs[3].sensor = image[3];

  //then individually update each block with collected inputs
  for(index = 0; index<LOGIC_BLOCKS; index++)
    {
      Logic_block_list[index].Update(CLB_inputs[index]);
    } //for
  for(index = 0; index<SENSOR_BLOCKS; index++)
    {
      Sensor_block_list[index].Update(CSB_inputs[index]);
    } //for
  //done --future iterations will propagate signals through more
  //than one block.
} //update state of array


void CROWBAR_class::Update(unsigned int gate_delays)
{
  //First check inputs
  //Then repeatedly call Update() routine
  //Stop early if convergence achieved
  bool converged = false;
  unsigned int counter = gate_delays;

  if(Check_Inputs())
    { 
      while((counter>0)&&(!converged))
	{
	  Update();
	  converged = Array_Converged();
	  counter--;
	} //while
      //cout<<"Array Converged: "<<gate_delays<<"\t"<<counter<<endl;
      //Warn if convergence not achieved after last iteration
      if(!converged)
	{
	  cerr<<"Warning: Signals still propagating"<<endl;
	} //if
    } //if
  else
    cerr<<"State not updated"<<endl;
} //Update


bool CROWBAR_class::Check_Inputs()
{
  bool okay_to_proceed = true;
  //Warn about stupid stuff (both clocks high, clk_b and reset high)
  voltage_type clk_a, clk_b, reset;

  clk_a = Pin_list[CLK_A].Get_Pad_Value();
  clk_b = Pin_list[CLK_B].Get_Pad_Value();
  reset = Pin_list[RST_L].Get_Pad_Value();

  switch(clk_a)
    {
    case High:
      if(clk_b == High)
	cerr<<"Warning: Both clocks asserted simultaneously"<<endl;
      break;
    case Low:
      break;
    case Floating:
    default:
      cerr<<"Warning: Unknown/Floating value for clk_a"<<endl;
    } //switch

  switch(clk_b)
    {
    case High:
      if(reset == Low) //reset is asserted low
	cerr<<"Warning: Both clk_b and reset asserted simultaneously"<<endl;
      break;
    case Low:
      break;
    case Floating:
    default:
      cerr<<"Warning: Unknown/Floating value for clk_b"<<endl;
    } //switch

  //also check for fatal errors (all vdd's must be high, and all gnd's low)
  unsigned int index;
  for(index = 0; index<PINS; index++)
    {
      if((Pin_list[index].Get_Type() == vdd)&&
	 (Pin_list[index].Get_Pad_Value() != High))
	{
	  okay_to_proceed = false;
	  cerr<<"Error: Vdd pin not connected to high voltage"<<endl;
	  break;
	} //if
      if((Pin_list[index].Get_Type() == gnd)&&
	 (Pin_list[index].Get_Pad_Value() != Low))
	{
	  okay_to_proceed = false;
	  cerr<<"Error: Gnd pin not connected to low voltage"<<endl;
	  break;
	} //if
    } //for
  return okay_to_proceed;
} //Check_Inputs


void CROWBAR_class::Apply_Voltage_To_Pin(pin_name_type pin_number, 
					voltage_type voltage)
{
  //Pin numbers in documentation range from 1 to 40
  //However, C++ likes to start counting with 0
  Pin_list[pin_number].Update_Pad_Value(voltage);
} //Apply_Voltage_To_Pin


void CROWBAR_class::Apply_Signal_To_Pin(pin_name_type pin_number, 
					signal_type signal)
{
  //Pin numbers in documentation range from 1 to 40
  //However, C++ likes to start counting with 0

  Pin_list[pin_number].Update_Internal_Value(signal);
} //Apply_Signal_To_Pin


void CROWBAR_class::Apply_Light_To_Sensor(unsigned int sensor_number, 
					   optical_type light)
{
  if(sensor_number < SENSOR_BLOCKS)
    {
      image[sensor_number] = light;
    } //if
  else
    cerr<<"Error: Sensor ID out of bounds"<<endl;
} //Apply_Signal_To_Sensor


voltage_type CROWBAR_class::Get_Pin_Voltage(pin_name_type pin_number)
{
  //Subtract one from pin_number to get index (0-39)
  return Pin_list[pin_number].Get_Pad_Value();
} //Get_Pin_Voltage



signal_type CROWBAR_class::Get_Pin_Signal(pin_name_type pin_number)
{
  return Pin_list[pin_number].Get_Internal_Value();
} //Get_Pin_Signal


void CROWBAR_class::Print_Scan()
{
  signal_type Sensor_flops[SENSOR_BLOCKS];
  signal_type Logic_flops[LOGIC_BLOCKS];

  //first get flop values (also known as test outputs for scan)
  unsigned int index;
  for(index = 0; index<SENSOR_BLOCKS; index++)
    {
      Sensor_flops[index] = Sensor_block_list[index].Get_Test_Output();
    } //for
  for(index = 0; index<LOGIC_BLOCKS; index++)
    {
      Logic_flops[index] = Logic_block_list[index].Get_Test_Output();
    } //for

  //print in order that they would come out of chip
  //S0,L2,S2,L6,L3,L0,L1,L4,L7,S3,L5,S1
  cout<<"Scan chain:"<<endl;
  cout<<signal_to_char(Sensor_flops[0]);
  cout<<signal_to_char(Logic_flops[2]);
  cout<<signal_to_char(Sensor_flops[2]);
  cout<<signal_to_char(Logic_flops[6]);
  cout<<signal_to_char(Logic_flops[3]);
  cout<<signal_to_char(Logic_flops[0]);
  cout<<signal_to_char(Logic_flops[1]);
  cout<<signal_to_char(Logic_flops[4]);
  cout<<signal_to_char(Logic_flops[7]);
  cout<<signal_to_char(Sensor_flops[3]);
  cout<<signal_to_char(Logic_flops[5]);
  cout<<signal_to_char(Sensor_flops[1]);
  cout<<endl;
} //Print_Scan


void CROWBAR_class::Print_Flop_Outputs()
{
  //For debugging, print outputs of flops in an array format
  signal_type Sensor_flops[SENSOR_BLOCKS];
  signal_type Logic_flops[LOGIC_BLOCKS];
  char Flop_Outputs[3][4];

  //first get flop values (also known as test outputs for scan)
  unsigned int index;
  for(index = 0; index<SENSOR_BLOCKS; index++)
    {
      Sensor_flops[index] = Sensor_block_list[index].Get_Test_Output();
    } //for
  for(index = 0; index<LOGIC_BLOCKS; index++)
    {
      Logic_flops[index] = Logic_block_list[index].Get_Test_Output();
    } //for

  //next get printable character representations
  Flop_Outputs[0][0] = signal_to_char(Sensor_flops[0]);
  Flop_Outputs[0][3] = signal_to_char(Sensor_flops[1]);
  Flop_Outputs[2][0] = signal_to_char(Sensor_flops[2]);
  Flop_Outputs[2][3] = signal_to_char(Sensor_flops[3]);
  Flop_Outputs[0][1] = signal_to_char(Logic_flops[0]);
  Flop_Outputs[0][2] = signal_to_char(Logic_flops[1]);
  Flop_Outputs[1][0] = signal_to_char(Logic_flops[2]);
  Flop_Outputs[1][1] = signal_to_char(Logic_flops[3]);
  Flop_Outputs[1][2] = signal_to_char(Logic_flops[4]);
  Flop_Outputs[1][3] = signal_to_char(Logic_flops[5]);
  Flop_Outputs[2][1] = signal_to_char(Logic_flops[6]);
  Flop_Outputs[2][2] = signal_to_char(Logic_flops[7]);

  //finally, print the characters to the screen
  cout<<"Flop outputs:"<<endl;
  cout<<"     "<<"        W        "<<endl;
  cout<<endl;
  cout<<"     ";
  for(index = 0; index<4; index++)
    {
      cout<<"----";
    } //for
  cout<<"-"<<endl;
  for(unsigned int index2 = 0; index2<3; index2++)
    {
      cout<<"     ";
      for(index = 0; index<4; index++)
	{
	  cout<<"|   ";
	} //for
      cout<<"|"<<endl;
      if(index2 != 1)
	cout<<"     ";
      else
	cout<<"S    ";
      for(index = 0; index<4; index++)
	{
	  cout<<"| "<<Flop_Outputs[index2][index]<<" ";
	} //for
      if(index2 != 1)
	cout<<"|"<<endl;
      else
	cout<<"|    N"<<endl;
      cout<<"     ";
      for(index = 0; index<4; index++)
	{
	  cout<<"|   ";
	} //for
      cout<<"|"<<endl;
      cout<<"     ";
      for(index = 0; index<4; index++)
	{
	  cout<<"----";
	} //for
      cout<<"-"<<endl;
    } //for
  cout<<endl;
  cout<<"     "<<"        E        "<<endl;
} //Print_Flop_Outputs


void CROWBAR_class::Print_Block_Outputs()
{
  //For debugging, print outputs of blocks in an array format
  //Similar to printing flop outputs (but we can print pins too!)
  signal_type Sensor_blocks[SENSOR_BLOCKS];
  signal_type Logic_blocks[LOGIC_BLOCKS];
  voltage_type West_in_pins[4]; //Top
  voltage_type East_in_pins[4]; //Bottom
  voltage_type South_in_pins[3]; //Left
  voltage_type North_in_pins[3]; //Right

  char Block_Outputs[3][4];
  char West_in[6];
  char East_in[6];
  char South_in[6];
  char North_in[6];

  //first get block values
  unsigned int index;
  for(index = 0; index<SENSOR_BLOCKS; index++)
    {
      Sensor_blocks[index] = Sensor_block_list[index].Get_Output();
    } //for
  for(index = 0; index<LOGIC_BLOCKS; index++)
    {
      Logic_blocks[index] = Logic_block_list[index].Get_Output();
    } //for
  
  //nasty-ness
  West_in_pins[0] = Pin_list[16].Get_Pad_Value(); //S0 in
  West_in_pins[1] = Pin_list[11].Get_Pad_Value(); //L0 in
  West_in_pins[2] = Pin_list[7].Get_Pad_Value(); //L1 in
  West_in_pins[3] = Pin_list[3].Get_Pad_Value(); //S1 in

  East_in_pins[0] = Pin_list[26].Get_Pad_Value(); //S2 in
  East_in_pins[1] = Pin_list[27].Get_Pad_Value(); //L6 in
  East_in_pins[2] = Pin_list[31].Get_Pad_Value(); //L7 in
  East_in_pins[3] = Pin_list[36].Get_Pad_Value(); //S3 in

  South_in_pins[0] = Pin_list[17].Get_Pad_Value(); //S0 in
  South_in_pins[1] = Pin_list[19].Get_Pad_Value(); //L2 in
  South_in_pins[2] = Pin_list[22].Get_Pad_Value(); //S2 in

  North_in_pins[0] = Pin_list[2].Get_Pad_Value(); //S1 in
  North_in_pins[1] = Pin_list[39].Get_Pad_Value(); //L5 in
  North_in_pins[2] = Pin_list[36].Get_Pad_Value(); //S3 in

  //next get printable character representations
  Block_Outputs[0][0] = signal_to_char(Sensor_blocks[0]);
  Block_Outputs[0][3] = signal_to_char(Sensor_blocks[1]);
  Block_Outputs[2][0] = signal_to_char(Sensor_blocks[2]);
  Block_Outputs[2][3] = signal_to_char(Sensor_blocks[3]);
  Block_Outputs[0][1] = signal_to_char(Logic_blocks[0]);
  Block_Outputs[0][2] = signal_to_char(Logic_blocks[1]);
  Block_Outputs[1][0] = signal_to_char(Logic_blocks[2]);
  Block_Outputs[1][1] = signal_to_char(Logic_blocks[3]);
  Block_Outputs[1][2] = signal_to_char(Logic_blocks[4]);
  Block_Outputs[1][3] = signal_to_char(Logic_blocks[5]);
  Block_Outputs[2][1] = signal_to_char(Logic_blocks[6]);
  Block_Outputs[2][2] = signal_to_char(Logic_blocks[7]);

  for(index = 0; index<4; index++)
    {
      West_in[index] = voltage_to_char(West_in_pins[index]);
      East_in[index] = voltage_to_char(East_in_pins[index]);
    } //for
  for(index = 0; index<3; index++)
    {
      South_in[index] = voltage_to_char(South_in_pins[index]);
      North_in[index] = voltage_to_char(North_in_pins[index]);
    } //for

  cout<<"Block outputs:"<<endl;
  cout<<"     "<<"        W        "<<endl;
  cout<<endl;
  cout<<"     "<<"  "<<West_in[0]<<"   "<<West_in[1]<<"   ";
  cout<<West_in[2]<<"   "<<West_in[3]<<endl;
  cout<<"     "<<"  "<<"|"<<"   "<<"|"<<"   "<<"|"<<"   "<<"|"<<endl;
  cout<<"     "<<"  "<<"V"<<"   "<<"V"<<"   "<<"V"<<"   "<<"V"<<endl;
  cout<<"     ";
  for(index = 0; index<4; index++)
    {
      cout<<"----";
    } //for
  cout<<"-"<<endl;
  for(unsigned int index2 = 0; index2<3; index2++)
    {
      cout<<"     ";
      for(index = 0; index<4; index++)
	{
	  cout<<"|   ";
	} //for
      cout<<"|"<<endl;
      if(index2 != 1)
	{
	  cout<<"  ";
	  cout<<South_in[index2]<<"->";
	}
      else
	{
	  cout<<"S ";
	  cout<<South_in[index2]<<"->";
	}
      for(index = 0; index<4; index++)
	{
	  cout<<"| "<<Block_Outputs[index2][index]<<" ";
	} //for
      if(index2 != 1)
	{
	  cout<<"|<-"<<North_in[index2]<<endl;
	}
      else
	{
	  cout<<"|<-"<<North_in[index2];
	  cout<<" N"<<endl;
	} //else
      cout<<"     ";
      for(index = 0; index<4; index++)
	{
	  cout<<"|   ";
	} //for
      cout<<"|"<<endl;
      cout<<"     ";
      for(index = 0; index<4; index++)
	{
	  cout<<"----";
	} //for
      cout<<"-"<<endl;
    } //for
  cout<<"     "<<"  "<<"^"<<"   "<<"^"<<"   "<<"^"<<"   "<<"^"<<endl;
  cout<<"     "<<"  "<<"|"<<"   "<<"|"<<"   "<<"|"<<"   "<<"|"<<endl;
  cout<<"     "<<"  "<<East_in[0]<<"   "<<East_in[1]<<"   ";
  cout<<East_in[2]<<"   "<<East_in[3]<<endl<<endl;
  cout<<"     "<<"        E        "<<endl;

} //Print_Block_Outputs


void CROWBAR_class::Print_Pinout()
{
  char pinout[] = "
            C
         P  O                       S  P
         A  N  L  L        L  L     C  A
         D  F  0  0        1  1     A  D
         |  O  O  W  C     O  W  R  N  |
         V  U  U  I  L     U  I  S  I  V
         D  T  T  N  K  G  T  N  T  N  D
         D  |  |  |  |  N  |  |  |  |  D
    +    2  H  H  H  A  D  H  H  L  H  1    +

          +------+------+------+------+
  S0WIN_H |      |      |      |      | S1WIN_H
  S0OUT_H | CSB0 | CLB0 | CLB1 | CSB1 | S1NIN_H
  S0SIN_H |      |      |      |      | S1OUT_H
          +------+------+------+------+
SCANOUT_H |      |      |      |      | L5NIN_H
  L2SIN_H | CLB2 | CLB3 | CLB4 | CLB5 | L5OUT_H
 L2SOUT_H |      |      |      |      |
          +------+------+------+------+
  S2OUT_H |      |      |      |      | S3NIN_H
  S2SIN_H | CSB2 | CLB6 | CLB7 | CSB3 | S3OUT_H
 CONFIN_H |      |      |      |      | S3EIN_H
          +------+------+------+------+

    -    P  T  S  L  L  V  C  L  L  C  P   -
         A  E  2  6  6  D  L  7  7  O  A
         D  S  E  E  O  D  K  E  O  N  D
         |  T  I  I  U     |  I  U  F  |
         G  |  N  N  T     B  N  T  O  G
         N  H  |  |  |        |  |  P  N
         D     H  H  H        H  H  T  D
         1                          |  2
                                    H
";
  cout<<"Pinout:"<<endl;
  cout<<pinout<<endl;
} //Print_Pinout()


bool CROWBAR_class::Array_Converged()
{
  bool converged = true; //initialized to true
  unsigned int index;
  for(index = 0; index<SENSOR_BLOCKS; index++)
    {
      converged = (converged && Sensor_block_list[index].Converged());
    } //for
  for(index = 0; index<LOGIC_BLOCKS; index++)
    {
      converged = (converged && Logic_block_list[index].Converged());
    } //for
  return converged;
} //Array_Converged


void CROWBAR_class::Print_Config()
{
  //handy for debugging
  //this is hard to understand...it is following the
  //path of the config chain. Next time, have one set
  //of config bits per block
  cout<<"Config:"<<endl;
  Print_CSB_Config_Bits(2);     cout<<' ';
  Print_CLB_LUT_Config_Bits(6); cout<<' ';
  Print_CLB_LUT_Config_Bits(7); cout<<' ';
  Print_CSB_Config_Bits(3);     cout<<' ';
  Print_CLB_Mux_Config_Bits(7); cout<<' ';
  Print_CLB_Mux_Config_Bits(6); cout<<' ';
  Print_CLB_LUT_Config_Bits(2); cout<<' ';
  Print_CLB_LUT_Config_Bits(3); cout<<' ';
  Print_CLB_LUT_Config_Bits(4); cout<<' ';
  Print_CLB_LUT_Config_Bits(5); cout<<' ';
  Print_CLB_Mux_Config_Bits(5); cout<<' ';
  Print_CLB_Mux_Config_Bits(4); cout<<' ';
  Print_CLB_Mux_Config_Bits(3); cout<<' ';
  Print_CLB_Mux_Config_Bits(2); cout<<' ';
  Print_CSB_Config_Bits(0);     cout<<' ';
  Print_CLB_LUT_Config_Bits(0); cout<<' ';
  Print_CLB_LUT_Config_Bits(1); cout<<' ';
  Print_CSB_Config_Bits(1);     cout<<' ';
  Print_CLB_Mux_Config_Bits(1); cout<<' ';
  Print_CLB_Mux_Config_Bits(0); cout<<endl;
} //Print_Config


void CROWBAR_class::Print_CSB_Config_Bits(unsigned int CSB_number)
{
  unsigned int index;
  signal_type bit;
  for(index = 0; index<Configurable_Mux8_Config_Bits; index++)
    {
      bit = Sensor_block_list[CSB_number].Get_Receiver_Config_Bit(index);
      cout<<signal_to_char(bit);
    } //for
  for(index = 0; index<Receiver_Config_Bits; index++)
    {
      bit = Sensor_block_list[CSB_number].Get_Receiver_Config_Bit(index);
      cout<<signal_to_char(bit);
    } //for
} //Print_CSB_Config_Bits


void CROWBAR_class::Print_CLB_LUT_Config_Bits(unsigned int CLB_number)
{
  unsigned int index;
  signal_type bit;
  for(index = 0; index<LUT8_Config_Bits; index++)
    {
      bit = Logic_block_list[CLB_number].Get_Lut_Config_Bit(index);
      cout<<signal_to_char(bit);
    } //for
} //Print_CLB_LUT_Config_Bits


void CROWBAR_class::Print_CLB_Mux_Config_Bits(unsigned int CLB_number)
{
  unsigned int index;
  signal_type bit;
  for(index = 0; index<Configurable_Mux4_Config_Bits; index++)
    {
      bit = Logic_block_list[CLB_number].Get_Mux0_Config_Bit(index);
      cout<<signal_to_char(bit);
    } //for
  for(index = 0; index<Configurable_Mux4_Config_Bits; index++)
    {
      bit = Logic_block_list[CLB_number].Get_Mux2_Config_Bit(index);
      cout<<signal_to_char(bit);
    } //for
  for(index = 0; index<Configurable_Mux4_Config_Bits; index++)
    {
      bit = Logic_block_list[CLB_number].Get_Mux1_Config_Bit(index);
      cout<<signal_to_char(bit);
    } //for
} //Print_CLB_Mux_Config_Bits


void CROWBAR_class::Print_A_Pin_Value(pin_name_type pin_name)
{
  char temp_label[MAX_STRING];
  voltage_type pad_voltage;
  signal_type internal_signal;
  pin_type type;

  Pin_Name_To_String(pin_name, temp_label);
  pad_voltage = Pin_list[pin_name].Get_Pad_Value();
  internal_signal = Pin_list[pin_name].Get_Internal_Value();
  type = Pin_list[pin_name].Get_Type();
  cout<<temp_label;
  if(strlen(temp_label) < (MAX_STRING-1)) //hack
    {
      for(unsigned int j = 0; j<((MAX_STRING-1)-strlen(temp_label)); j++)
	cout<<' ';
    } //if
  cout<<"\t(";
  Print_pin_type(type);
  cout<<")\t"<<"voltage: "<<voltage_to_char(pad_voltage)<<", ";
  cout<<"internal: "<<signal_to_char(internal_signal)<<endl;
} //Print_A_Pin_Value


void CROWBAR_class::Print_Pin_Values()
{
  unsigned int index;

  cout<<"Pins:"<<endl;
  for(index = 0; index<PINS; index++)
    {
      cout<<index<<":\t";
      Print_A_Pin_Value(index);
    } //for
} //Print_Pin_Values


void CROWBAR_class::Print_Image()
{
  //For debugging, print inputs to sensors in an array format
  unsigned int index;
  cout<<"Sensor inputs:"<<endl;
  cout<<"     "<<"        W        "<<endl;
  cout<<endl;
  cout<<"     ";
  for(index = 0; index<4; index++)
    {
      cout<<"----";
    } //for
  cout<<"-"<<endl;
  for(unsigned int index2 = 0; index2<3; index2++)
    {
      cout<<"     ";
      for(index = 0; index<4; index++)
	{
	  cout<<"|   ";
	} //for
      cout<<"|"<<endl;
      if(index2 != 1)
	cout<<"     ";
      else
	cout<<"S    ";
      for(index = 0; index<4; index++)
	{
	  cout<<"| ";
	  switch((index2*4)+index)
	    {
	    case 0:
	      cout<<optical_to_char(image[0]);
	      break;
	    case 3:
	      cout<<optical_to_char(image[1]);
	      break;
	    case 8:
	      cout<<optical_to_char(image[2]);
	      break;
	    case 11:
	      cout<<optical_to_char(image[3]);
	      break;
	    default:
	      cout<<" ";
	    } //switch
	  cout<<" ";
	} //for
      if(index2 != 1)
	cout<<"|"<<endl;
      else
	cout<<"|    N"<<endl;
      cout<<"     ";
      for(index = 0; index<4; index++)
	{
	  cout<<"|   ";
	} //for
      cout<<"|"<<endl;
      cout<<"     ";
      for(index = 0; index<4; index++)
	{
	  cout<<"----";
	} //for
      cout<<"-"<<endl;
    } //for
  cout<<endl;
  cout<<"     "<<"        E        "<<endl;
} //Print_Image

///////////////////////////////////////////////////////////////////////////////
//end implementation file

